==> 0.8.4.tar.gz: lib/src/mirrors/_.dart <==
part of fields.mirrors;

final _iterableClassMirror = reflectClass(Iterable);
final _listClassMirror = reflectClass(List);
final _setClassMirror = reflectClass(Set);
final _mapClassMirror = reflectClass(Map);
final _typedDataClassMirror = reflectClass(TypedData);
final _objectClassMirror = reflectClass(Object);
final _dynamicTypeMirror = currentMirrorSystem().dynamicType;
final _defaultEditorClassMirror = reflectClass(DefaultEditor);

Type _typeOf(TypeMirror tm) {
  if (tm == _dynamicTypeMirror || tm is! ClassMirror) {
    return dynamic;
  } else {
    return (tm as ClassMirror).reflectedType;
  }
}

const List<Type> _atomicTypes = const [ 
    bool, double, int, List, Map, Set, Object, num, String, Type, TypedData 
];

/// A list of types that are atomic from the viewpoint of this library and its
/// reflection.
/// 
/// Some user-defined types may lead to a [StackOverflowError] when this library
/// performs reflection of instance members.  For example, when a class declares
/// a getter that allocates a new object of its own class on each call.  This 
/// list provides a way to prevent reflective invocation of such getters.
/// 
/// This list is initialized with a default list that contains simple types from
/// `dart:core`; it may be modified to register user-defined atomic types.
final List<Type> atomicTypes = new List.from(_atomicTypes);

bool _isAtomicType(Type t) {
  if (atomicTypes.contains(t)) {
    return true;      
  } else {
    if (dynamic == t) {
      return true;
    } else {
      ClassMirror cm = reflectClass(t);
      if (_isAssignableFrom(_iterableClassMirror, cm) ||
          _isAssignableFrom(_listClassMirror, cm) ||
          _isAssignableFrom(_mapClassMirror, cm) ||
          _isAssignableFrom(_setClassMirror, cm) ||
          _isAssignableFrom(_typedDataClassMirror, cm)) {
        return true;
      }
    }
    return false;
  }
}

/// Returns `true` of the given [type] and the given [other] represent the same
/// type, or if the [type] is in the inheritance hierarchy of [other], or if
/// [other] implements the interface of [type]; else returns `false`.
bool _isAssignableFrom(TypeMirror type, TypeMirror other) {
  assert(type != null);
  assert(other != null);
  if (type == _dynamicTypeMirror) return true;
  if (!other.isOriginalDeclaration) other = other.originalDeclaration;  
  if (other == type) return true;
  if (other is! ClassMirror) return false;
  ClassMirror otherCm = other;  
  if (otherCm.superclass == null) return type == _objectClassMirror;
  if (otherCm.superclass == type) return true;
  if (otherCm.superinterfaces.any((i) => i == type)) return true;
  return _isAssignableFrom(type, otherCm.superclass);
}

Map _parseMetadata(DeclarationMirror declaration) {
  assert(declaration != null);
  final metadata = new Map()        
  ..[FieldEditorType] = null
  ..[UseValueTransform] = null
  ..[ReadOnly] = false
  ..[Ignore] = false  
  ..[DisplayName] = null
  ..[Description] = null
  ..[FieldAnnotation] = <FieldAnnotation>[];
  declaration.metadata.forEach((InstanceMirror mirror) {
    final instance = mirror.reflectee;
    if (instance is FieldEditorType) {
      // Disallow > 1 @FieldEditorType
      assert(metadata[FieldEditorType] == null);
      // TODO(rms): validate that the editor type isAssignableTo FieldEditor.
      metadata[FieldEditorType] = instance.editorType;
    } else if (instance is UseValueTransform) {
      // Disallow > 1 @UseValueTransform
      assert(metadata[UseValueTransform] == null);
      metadata[UseValueTransform] = instance.transform;
    } else if (instance == ReadOnly) {
      metadata[ReadOnly] = true;
    } else if (instance == Ignore) {
      metadata[Ignore] = true;
    } else if (instance is DisplayName) {
      // Disallow > 1 @DisplayName
      assert(metadata[DisplayName] == null);
      metadata[DisplayName] = instance.name;
    } else if (instance is Description) {
      // Disallow > 1 @Description
      assert(metadata[Description] == null);
      metadata[Description] = instance.description;
    } else if (instance is FieldAnnotation) {
      // Editor-specific annotations, store them in the field as-is.
      metadata[FieldAnnotation].add(instance);
    }
  });
  return metadata;
}

List<Field> reflectFields(Object object, [Field parent]) {
  final fields = new List<Field>();  
  // Prefer the parent's type over the object's runtimeType when available.
  final type = parent == null ? object.runtimeType : parent.type;
  // We don't reflect `atomicTypes`
  if (_isAtomicType(type)) {
    return fields;
  }
  final instanceMirror = reflect(object);
  // We reflect the ClassMirror from `Object.runtimeType` instead of getting
  // the ClassMirror from the `InstanceMirror.type` to support proxy objects.
  final classMirror = reflectClass(object.runtimeType);
  // TODO: replace `declarations` with `instanceMembers` pending:
  // https://code.google.com/p/dart/issues/detail?id=18933
  classMirror.declarations.forEach((key, declaration) {
    if (declaration is MethodMirror && 
        declaration.isGetter && 
        !declaration.isStatic) {
      final field = new FieldMirror.forGetter(
          key,
          instanceMirror, 
          getter: declaration, 
          parent: parent);
      if (field != null) {
        fields.add(field);
      }        
    } else if (declaration is VariableMirror && !declaration.isStatic) {
      final field = new FieldMirror.forVariable(
          key,
          instanceMirror, 
          variable: declaration,
          parent: parent);
      if (field != null) {
        fields.add(field);
      }
    }
  });
  return fields;
}

/// Creates a new field editor for the given [field] using mirror reflection.
/// 
/// The class of the created field editor is the first of the following that 
/// matches:
/// 
/// * The class given by [Field.editorType] if it is not `null`.
/// * The registered editor class for the [Field.type], if any.
/// * The [DefaultEditor] class.
/// 
/// If none of the above conditions match then an error is thrown.
FieldEditor reflectFieldEditor(Field field) {
  ClassMirror cm = _defaultEditorClassMirror;
  if (field.editorType != null) {
    cm = reflectClass(field.editorType);
  } else if (fieldEditors.containsKey(field.type)) {
    cm = reflectClass(fieldEditors[field.type]);
  }
  assert(cm != null);
  return cm.newInstance(const Symbol(''),[field]).reflectee;
}

==> 0.8.11.tar.gz: lib/src/mirrors/_.dart <==
part of fields.mirrors;

final _iterableClassMirror = reflectClass(Iterable);
final _listClassMirror = reflectClass(List);
final _setClassMirror = reflectClass(Set);
final _mapClassMirror = reflectClass(Map);
final _typedDataClassMirror = reflectClass(TypedData);
final _objectClassMirror = reflectClass(Object);
final _dynamicTypeMirror = currentMirrorSystem().dynamicType;
final _defaultEditorClassMirror = reflectClass(DefaultEditor);

Type _typeOf(TypeMirror tm) {
  if (tm == _dynamicTypeMirror || tm is! ClassMirror) {
    return dynamic;
  } else {
    return (tm as ClassMirror).reflectedType;
  }
}

const List<Type> _atomicTypes = const [ 
    bool, double, int, List, Map, Set, Object, num, String, Type, TypedData 
];

/// A list of types that are atomic from the viewpoint of this library and its
/// reflection.
/// 
/// Some user-defined types may lead to a [StackOverflowError] when this library
/// performs reflection of instance members.  For example, when a class declares
/// a getter that allocates a new object of its own class on each call.  This 
/// list provides a way to prevent reflective invocation of such getters.
/// 
/// This list is initialized with a default list that contains simple types from
/// `dart:core`; it may be modified to register user-defined atomic types.
final List<Type> atomicTypes = new List.from(_atomicTypes);

bool _isAtomicType(Type t) {
  if (atomicTypes.contains(t)) {
    return true;      
  } else {
    if (dynamic == t) {
      return true;
    } else {
      ClassMirror cm = reflectClass(t);
      if (_isAssignableFrom(_iterableClassMirror, cm) ||
          _isAssignableFrom(_listClassMirror, cm) ||
          _isAssignableFrom(_mapClassMirror, cm) ||
          _isAssignableFrom(_setClassMirror, cm) ||
          _isAssignableFrom(_typedDataClassMirror, cm)) {
        return true;
      }
    }
    return false;
  }
}

/// Returns `true` of the given [type] and the given [other] represent the same
/// type, or if the [type] is in the inheritance hierarchy of [other], or if
/// [other] implements the interface of [type]; else returns `false`.
bool _isAssignableFrom(TypeMirror type, TypeMirror other) {
  assert(type != null);
  assert(other != null);
  if (type == _dynamicTypeMirror) return true;
  if (!other.isOriginalDeclaration) other = other.originalDeclaration;  
  if (other == type) return true;
  if (other is! ClassMirror) return false;
  ClassMirror otherCm = other;  
  if (otherCm.superclass == null) return type == _objectClassMirror;
  if (otherCm.superclass == type) return true;
  if (otherCm.superinterfaces.any((i) => i == type)) return true;
  return _isAssignableFrom(type, otherCm.superclass);
}

Map _parseMetadata(DeclarationMirror declaration) {
  assert(declaration != null);
  final metadata = new Map()        
  ..[FieldEditorType] = null
  ..[UseValueTransform] = null
  ..[ReadOnly] = false
  ..[Ignore] = false  
  ..[DisplayName] = null
  ..[Description] = null
  ..[FieldAnnotation] = <FieldAnnotation>[];
  declaration.metadata.forEach((InstanceMirror mirror) {
    final instance = mirror.reflectee;
    if (instance is FieldEditorType) {
      // Disallow > 1 @FieldEditorType
      assert(metadata[FieldEditorType] == null);
      // TODO: validate that the editor type isAssignableTo FieldEditor.
      metadata[FieldEditorType] = instance.editorType;
    } else if (instance is UseValueTransform) {
      // Disallow > 1 @UseValueTransform
      assert(metadata[UseValueTransform] == null);
      metadata[UseValueTransform] = instance.transform;
    } else if (instance == ReadOnly) {
      metadata[ReadOnly] = true;
    } else if (instance == Ignore) {
      metadata[Ignore] = true;
    } else if (instance is DisplayName) {
      // Disallow > 1 @DisplayName
      assert(metadata[DisplayName] == null);
      metadata[DisplayName] = instance.name;
    } else if (instance is Description) {
      // Disallow > 1 @Description
      assert(metadata[Description] == null);
      metadata[Description] = instance.description;
    } else if (instance is FieldAnnotation) {
      // Editor-specific annotations, store them in the field as-is.
      metadata[FieldAnnotation].add(instance);
    }
  });
  return metadata;
}

List<Field> reflectFields(Object object, [Field parent]) {
  final fields = new List<Field>();  
  // Prefer the parent's type over the object's runtimeType when available.
  final type = parent == null ? object.runtimeType : parent.type;
  // We don't reflect `atomicTypes`
  if (_isAtomicType(type)) {
    return fields;
  }
  final instanceMirror = reflect(object);
  // We reflect the ClassMirror from `Object.runtimeType` instead of getting
  // the ClassMirror from the `InstanceMirror.type` to support proxy objects.
  final classMirror = reflectClass(object.runtimeType);
  // TODO: replace `declarations` with `instanceMembers` pending:
  // https://code.google.com/p/dart/issues/detail?id=18933
  classMirror.declarations.forEach((key, declaration) {
    if (declaration is MethodMirror && 
        declaration.isGetter && 
        !declaration.isStatic) {
      final field = new FieldMirror.forGetter(
          key,
          instanceMirror, 
          getter: declaration, 
          parent: parent);
      if (field != null) {
        fields.add(field);
      }        
    } else if (declaration is VariableMirror && !declaration.isStatic) {
      final field = new FieldMirror.forVariable(
          key,
          instanceMirror, 
          variable: declaration,
          parent: parent);
      if (field != null) {
        fields.add(field);
      }
    }
  });
  return fields;
}

/// Creates a new field editor for the given [field] using mirror reflection.
/// 
/// The class of the created field editor is the first of the following that 
/// matches:
/// 
/// * The class given by [Field.editorType] if it is not `null`.
/// * The registered editor class for the [Field.type], if any.
/// * The [DefaultEditor] class.
/// 
/// If none of the above conditions match then an error is thrown.
FieldEditor reflectFieldEditor(Field field) {
  ClassMirror cm = _defaultEditorClassMirror;
  if (field.editorType != null) {
    cm = reflectClass(field.editorType);
  } else if (fieldEditors.containsKey(field.type)) {
    cm = reflectClass(fieldEditors[field.type]);
  }
  assert(cm != null);
  return cm.newInstance(const Symbol(''),[field]).reflectee;
}
